# -*- coding: utf-8 -*-
# 版权所有 2019 深圳米筐科技有限公司（下称“米筐科技”）
#
# 除非遵守当前许可，否则不得使用本软件。
#
#     * 非商业用途（非商业用途指个人出于非商业目的使用本软件，或者高校、研究所等非营利机构出于教育、科研等目的使用本软件）：
#         遵守 Apache License 2.0（下称“Apache 2.0 许可”），
#         您可以在以下位置获得 Apache 2.0 许可的副本：http://www.apache.org/licenses/LICENSE-2.0。
#         除非法律有要求或以书面形式达成协议，否则本软件分发时需保持当前许可“原样”不变，且不得附加任何条件。
#
#     * 商业用途（商业用途指个人出于任何商业目的使用本软件，或者法人或其他组织出于任何目的使用本软件）：
#         未经米筐科技授权，任何个人不得出于任何商业目的使用本软件（包括但不限于向第三方提供、销售、出租、出借、转让本软件、
#         本软件的衍生产品、引用或借鉴了本软件功能或源代码的产品或服务），任何法人或其他组织不得出于任何目的使用本软件，
#         否则米筐科技有权追究相应的知识产权侵权责任。
#         在此前提下，对本软件的使用同样需要遵守 Apache 2.0 许可，Apache 2.0 许可与本许可冲突之处，以本许可为准。
#         详细的授权流程，请联系 public@ricequant.com 获取。

from copy import copy
from datetime import datetime

from rqalpha.core.events import EVENT, Event
from rqalpha.utils.rq_json import convert_dict_to_json, convert_json_to_dict
from rqalpha.utils.logger import system_log


class Executor(object):
    def __init__(self, env):
        self._env = env
        self._last_before_trading = None

    def get_state(self):
        return convert_dict_to_json({"last_before_trading": self._last_before_trading}).encode('utf-8')

    def set_state(self, state):
        self._last_before_trading = convert_json_to_dict(state.decode('utf-8')).get("last_before_trading")

    def run(self, bar_dict):
        conf = self._env.config.base
        for event in self._env.event_source.events(conf.start_date, conf.end_date, conf.frequency):
            if event.event_type == EVENT.TICK:
                if self._ensure_before_trading(event):
                    self._split_and_publish(event)
            elif event.event_type == EVENT.BAR:
                if self._ensure_before_trading(event):
                    bar_dict.update_dt(event.calendar_dt)
                    event.bar_dict = bar_dict
                    self._split_and_publish(event)
            elif event.event_type == EVENT.OPEN_AUCTION:
                if self._ensure_before_trading(event):
                    bar_dict.update_dt(event.calendar_dt)
                    event.bar_dict = bar_dict
                    self._split_and_publish(event)
            elif event.event_type == EVENT.BEFORE_TRADING:
                self._ensure_before_trading(event)
            elif event.event_type == EVENT.AFTER_TRADING:
                self._split_and_publish(event)
            else:
                self._env.event_bus.publish_event(event)

        # publish settlement after last day
        if self._env.trading_dt.date() == conf.end_date:
            self._split_and_publish(Event(EVENT.SETTLEMENT))

    def _ensure_before_trading(self, event):
        # return True if before_trading won't run this time
        if self._last_before_trading == event.trading_dt.date() or self._env.config.extra.is_hold:
            return True
        if self._last_before_trading:
            # don't publish settlement on first day
            previous_trading_date = self._env.data_proxy.get_previous_trading_date(event.trading_dt).date()
            if self._env.trading_dt.date() != previous_trading_date:
                self._env.update_time(
                    datetime.combine(previous_trading_date, self._env.calendar_dt.time()),
                    datetime.combine(previous_trading_date, self._env.trading_dt.time())
                )
            system_log.debug("publish settlement events with calendar_dt={}, trading_dt={}".format(
                self._env.calendar_dt, self._env.trading_dt
            ))
            self._split_and_publish(Event(EVENT.SETTLEMENT))
        self._last_before_trading = event.trading_dt.date()
        self._split_and_publish(Event(EVENT.BEFORE_TRADING, calendar_dt=event.calendar_dt, trading_dt=event.trading_dt))
        return False

    EVENT_SPLIT_MAP = {
        EVENT.BEFORE_TRADING: (EVENT.PRE_BEFORE_TRADING, EVENT.BEFORE_TRADING, EVENT.POST_BEFORE_TRADING),
        EVENT.BAR: (EVENT.PRE_BAR, EVENT.BAR, EVENT.POST_BAR),
        EVENT.TICK: (EVENT.PRE_TICK, EVENT.TICK, EVENT.POST_TICK),
        EVENT.AFTER_TRADING: (EVENT.PRE_AFTER_TRADING, EVENT.AFTER_TRADING, EVENT.POST_AFTER_TRADING),
        EVENT.SETTLEMENT: (EVENT.PRE_SETTLEMENT, EVENT.SETTLEMENT, EVENT.POST_SETTLEMENT),
        EVENT.OPEN_AUCTION: (EVENT.PRE_OPEN_AUCTION, EVENT.OPEN_AUCTION, EVENT.POST_OPEN_AUCTION),
    }

    def _split_and_publish(self, event):
        if hasattr(event, "calendar_dt") and hasattr(event, "trading_dt"):
            self._env.update_time(event.calendar_dt, event.trading_dt)
        for event_type in self.EVENT_SPLIT_MAP[event.event_type]:
            e = copy(event)
            e.event_type = event_type
            self._env.event_bus.publish_event(e)



# -*- coding: utf-8 -*-
# 版权所有 2019 深圳米筐科技有限公司（下称“米筐科技”）
#
# 除非遵守当前许可，否则不得使用本软件。
#
#     * 非商业用途（非商业用途指个人出于非商业目的使用本软件，或者高校、研究所等非营利机构出于教育、科研等目的使用本软件）：
#         遵守 Apache License 2.0（下称“Apache 2.0 许可”），
#         您可以在以下位置获得 Apache 2.0 许可的副本：http://www.apache.org/licenses/LICENSE-2.0。
#         除非法律有要求或以书面形式达成协议，否则本软件分发时需保持当前许可“原样”不变，且不得附加任何条件。
#
#     * 商业用途（商业用途指个人出于任何商业目的使用本软件，或者法人或其他组织出于任何目的使用本软件）：
#         未经米筐科技授权，任何个人不得出于任何商业目的使用本软件（包括但不限于向第三方提供、销售、出租、出借、转让本软件、
#         本软件的衍生产品、引用或借鉴了本软件功能或源代码的产品或服务），任何法人或其他组织不得出于任何目的使用本软件，
#         否则米筐科技有权追究相应的知识产权侵权责任。
#         在此前提下，对本软件的使用同样需要遵守 Apache 2.0 许可，Apache 2.0 许可与本许可冲突之处，以本许可为准。
#         详细的授权流程，请联系 public@ricequant.com 获取。

from collections import UserDict, deque
from datetime import date
from decimal import Decimal
from typing import Dict, Iterable, Tuple, Optional, Deque, List

from rqalpha.const import POSITION_DIRECTION, POSITION_EFFECT
from rqalpha.environment import Environment
from rqalpha.interface import AbstractPosition
from rqalpha.model.instrument import Instrument
from rqalpha.model.order import Order
from rqalpha.model.trade import Trade
from rqalpha.utils import is_valid_price
from rqalpha.utils.i18n import gettext as _
from rqalpha.utils.repr import PropertyReprMeta, property_repr
from rqalpha.utils.logger import user_system_log


def new_position_meta():
    type_map = {}

    class Meta(PropertyReprMeta):
        def __new__(mcs, *args, **kwargs):
            cls = super(Meta, mcs).__new__(mcs, *args, **kwargs)
            try:
                instrument_types = cls.__instrument_types__
            except AttributeError:
                pass
            else:
                for instrument_type in instrument_types:
                    type_map[instrument_type] = cls
            return cls

    return type_map, Meta


POSITION_TYPE_MAP, PositionMeta = new_position_meta()
POSITION_PROXY_TYPE_MAP, PositionProxyMeta = new_position_meta()


class Position(AbstractPosition, metaclass=PositionMeta):
    __repr_properties__ = (
        "order_book_id", "direction", "quantity", "market_value", "trading_pnl", "position_pnl", "last_price"
    )

    # 用于注册该 Position 类型适用的 instrument_type
    __instrument_types__ = []

    def __new__(cls, order_book_id, direction, init_quantity=0, init_price=None):
        if cls == Position:
            ins_type = Environment.get_instance().data_proxy.instrument(order_book_id).type
            try:
                position_cls = POSITION_TYPE_MAP[ins_type]
            except KeyError:
                raise NotImplementedError("")
            return position_cls.__new__(position_cls, order_book_id, direction, init_quantity, init_price)
        else:
            return object.__new__(cls)

    def __init__(self, order_book_id, direction, init_quantity=0, init_price=None):
        self._env = Environment.get_instance()

        self._order_book_id = order_book_id
        self._instrument = self._env.data_proxy.instrument(order_book_id)  # type: Instrument
        self._direction = direction

        self._quantity = init_quantity
        self._old_quantity = init_quantity
        self._logical_old_quantity = 0

        self._avg_price: float = init_price or 0
        self._trade_cost: float = 0
        self._transaction_cost: float = 0
        self._prev_close: Optional[float] = init_price
        self._last_price: Optional[float] = init_price

        self._direction_factor = 1 if direction == POSITION_DIRECTION.LONG else -1

        self._queue = PositionQueue()
        if init_quantity:
            self._queue.handle_trade(init_quantity, self._env.trading_dt.date())

    @property
    def order_book_id(self):
        # type: () -> str
        return self._order_book_id

    @property
    def direction(self):
        # type: () -> POSITION_DIRECTION
        return self._direction

    @property
    def quantity(self):
        # type: () -> int
        return self._quantity

    @property
    def transaction_cost(self):
        # type: () -> float
        return self._transaction_cost

    @property
    def avg_price(self):
        # type: () -> float
        return self._avg_price

    @property
    def trading_pnl(self):
        # type: () -> float
        trade_quantity = self._quantity - self._logical_old_quantity
        return (trade_quantity * self.last_price - self._trade_cost) * self._direction_factor

    @property
    def position_pnl(self):
        # type: () -> float
        if self._logical_old_quantity:
            return self._logical_old_quantity * (self.last_price - self.prev_close) * self._direction_factor
        else:
            return 0

    @property
    def pnl(self):
        # type: () -> float
        """
        返回该持仓的累积盈亏
        """
        return (self.last_price - self.avg_price) * self._quantity * self._direction_factor

    @property
    def market_value(self):
        # type: () -> float
        return self.last_price * self._quantity if self._quantity else 0

    @property
    def equity(self):
        # type: () -> float
        return self.last_price * self._quantity if self._quantity else 0

    @property
    def prev_close(self):
        # type: () -> float
        if not is_valid_price(self._prev_close):
            self._prev_close = self._env.data_proxy.get_prev_close(self._order_book_id, self._env.trading_dt)
            if not is_valid_price(self._prev_close):
                user_system_log.warn("invalid prev close of {}: {}".format(self._order_book_id, self._prev_close))
        return self._prev_close

    @property
    def last_price(self):
        # type: () -> float
        if not is_valid_price(self._last_price):
            self._last_price = self._env.data_proxy.get_last_price(self._order_book_id)
            if not is_valid_price(self._last_price):
                user_system_log.warn("invalid last price of {}: {}".format(self._order_book_id, self._last_price))
        return self._last_price

    @property
    def closable(self):
        # type: () -> int
        """
        可平仓位
        """
        order_quantity = sum(o.unfilled_quantity for o in self._open_orders if o.position_effect in (
            POSITION_EFFECT.CLOSE, POSITION_EFFECT.CLOSE_TODAY, POSITION_EFFECT.EXERCISE
        ))
        return self._quantity - order_quantity

    @property
    def today_closable(self):
        # type: () -> int
        return self._quantity - self._old_quantity - sum(
            o.unfilled_quantity for o in self._open_orders if o.position_effect == POSITION_EFFECT.CLOSE_TODAY
        )

    def get_state(self):
        """"""
        return {
            "old_quantity": self._old_quantity,
            "logical_old_quantity": self._logical_old_quantity,
            "quantity": self._quantity,
            "avg_price": self._avg_price,
            "trade_cost": self._trade_cost,
            "transaction_cost": self._transaction_cost,
            "prev_close": self._prev_close,
            "position_queue": list(self._queue.queue)
        }

    def set_state(self, state):
        """"""
        self._old_quantity = state.get("old_quantity", 0)
        self._logical_old_quantity = state.get("logical_old_quantity", self._old_quantity)
        if "quantity" in state:
            self._quantity = state["quantity"]
        else:
            # forward compatible
            self._quantity = self._old_quantity + state.get("today_quantity", 0)
        self._avg_price = state.get("avg_price", 0)
        self._trade_cost = state.get("trade_cost", 0)
        self._transaction_cost = state.get("transaction_cost", 0)
        self._prev_close = state.get("prev_close")
        self._queue.set_sate(state.get("position_queue", []))

    def before_trading(self, trading_date):
        # type: (date) -> float
        # 返回该阶段导致总资金的变化量
        self._old_quantity = self._quantity
        self._logical_old_quantity = self._old_quantity
        self._trade_cost = self._non_closable = 0
        self._transaction_cost = 0
        self._prev_close = None
        return 0

    def apply_trade(self, trade):
        # type: (Trade) -> float
        # 返回总资金的变化量
        self._transaction_cost += trade.transaction_cost
        if trade.position_effect == POSITION_EFFECT.OPEN:
            self._queue.handle_trade(trade.last_quantity, self._env.trading_dt.date())
            if self._quantity < 0:
                self._avg_price = trade.last_price if self._quantity + trade.last_quantity > 0 else 0
            else:
                cost = self._quantity * self._avg_price + trade.last_quantity * trade.last_price
                self._avg_price = cost / (self._quantity + trade.last_quantity)
            self._quantity += trade.last_quantity
            self._trade_cost += trade.last_price * trade.last_quantity
            return (-1 * trade.last_price * trade.last_quantity) - trade.transaction_cost
        elif trade.position_effect == POSITION_EFFECT.CLOSE:
            # 先平昨，后平今
            self._queue.handle_trade(-trade.last_quantity, self._env.trading_dt.date())
            self._old_quantity -= min(trade.last_quantity, self._old_quantity)
            self._quantity -= trade.last_quantity
            self._trade_cost -= trade.last_price * trade.last_quantity
            return trade.last_price * trade.last_quantity - trade.transaction_cost
        else:
            raise NotImplementedError("{} does not support position effect {}".format(
                self.__class__.__name__, trade.position_effect
            ))

    def settlement(self, trading_date):
        # type: (date) -> float
        # 返回该阶段导致总资金的变化量以及反映该阶段引起其他持仓变化的虚拟交易，虚拟交易用于换代码，转股等操作
        return 0

    def update_last_price(self, price):
        self._last_price = price

    def calc_close_today_amount(self, trade_amount, position_effect):
        return 0

    @property
    def position_queue(self):
        # type: () -> deque[tuple[date, int]]
        """
        获取持仓队列，返回每笔持仓的开仓时间和数量
        格式为 [(开仓日期, 持仓数量), ...]
        """
        return self._queue.queue

    @property
    def _open_orders(self):
        # type: () -> Iterable[Order]
        for order in self._env.broker.get_open_orders(self.order_book_id):
            if order.position_direction == self._direction:
                yield order


class PositionQueue:
    def __init__(self):
        self._queue: Deque[Tuple[date, int]] = deque()

    def set_sate(self, state: List[Tuple[date, int]]):
        self._queue = deque(state)

    def handle_trade(self, delta_quantity, trading_date: date, close_today: bool = False):
        if delta_quantity == 0:
            return
        if self._queue and self._queue[0][1] * delta_quantity < 0:
            # 从正平仓到负或者从负平仓到正
            if close_today:
                # 强制先平今
                d, qty_in_queue = self._queue[-1]
                if d >= trading_date:
                    if abs(qty_in_queue) <= abs(delta_quantity):
                        delta_quantity += qty_in_queue
                        self._queue.pop()
                    else:
                        self._queue[-1] = (d, qty_in_queue + delta_quantity)
                        delta_quantity = 0
            while delta_quantity and self._queue:
                d, qty_in_queue = self._queue[0]
                if abs(qty_in_queue) <= abs(delta_quantity):
                    # 当前持仓批次全部平掉
                    delta_quantity += qty_in_queue
                    self._queue.popleft()
                else:
                    # 当前持仓批次部分平掉
                    self._queue[0] = (d, qty_in_queue + delta_quantity)
                    delta_quantity = 0
            if delta_quantity != 0:
                # queue 消耗光了，delta_quantity 还没消耗完
                self._queue.append((trading_date, delta_quantity))
        else:
            # 开仓
            if self._queue and self._queue[-1][0] == trading_date:
                # 当日已有开仓，更新数量
                _, quantity = self._queue.pop()
                self._queue.append((trading_date, quantity + delta_quantity))
            else:
                # 当日首次开仓
                self._queue.append((trading_date, delta_quantity))

    def handle_split(self, ratio: Decimal, expected_quantity):
        self._queue = deque(
            (d, round(Decimal(q) * ratio)) for d, q in self._queue
        )
        diff = expected_quantity - sum(q for _, q in self._queue)
        if diff != 0:
            # 有可能因为 round 的原因导致差1股
            self._queue[-1] = (self._queue[-1][0], self._queue[-1][1] + diff)


    def clear(self):
        self._queue.clear()

    @property
    def queue(self):
        return self._queue.copy()


class PositionProxy(metaclass=PositionProxyMeta):
    __repr_properties__ = (
        "order_book_id", "positions"
    )
    # 用于注册该 Position 类型适用的 instrument_type
    __instrument_types__ = []

    def __init__(self, long, short):
        # type: (Position, Position) -> PositionProxy
        self._long = long
        self._short = short

    __repr__ = property_repr

    @property
    def type(self):
        raise NotImplementedError

    @property
    def order_book_id(self):
        return self._long.order_book_id

    @property
    def last_price(self):
        return self._long.last_price

    @property
    def market_value(self):
        return self._long.market_value - self._short.market_value

    # -- PNL 相关
    @property
    def position_pnl(self):
        """
        [float] 昨仓盈亏，当前交易日盈亏中来源于昨仓的部分

        多方向昨仓盈亏 = 昨日收盘时的持仓 * 合约乘数 * (最新价 - 昨收价)
        空方向昨仓盈亏 = 昨日收盘时的持仓 * 合约乘数 * (昨收价 - 最新价)

        """
        return self._long.position_pnl + self._short.position_pnl

    @property
    def trading_pnl(self):
        """
        [float] 交易盈亏，当前交易日盈亏中来源于当日成交的部分

        单比买方向成交的交易盈亏 = 成交量 * (最新价 - 成交价)
        单比卖方向成交的交易盈亏 = 成交量 * (成交价 - 最新价)

        """
        return self._long.trading_pnl + self._short.trading_pnl

    @property
    def daily_pnl(self):
        """
        [float] 当日盈亏

        当日盈亏 = 昨仓盈亏 + 交易盈亏

        """
        return self._long.position_pnl + self._long.trading_pnl + self._short.position_pnl + \
               self._short.trading_pnl - self.transaction_cost

    @property
    def pnl(self):
        return self._long.pnl + self._short.pnl

    # -- Quantity 相关
    @property
    def open_orders(self):
        return Environment.get_instance().broker.get_open_orders(self.order_book_id)

    # -- Margin 相关
    @property
    def margin(self):
        """
        [float] 保证金

        保证金 = 持仓量 * 最新价 * 合约乘数 * 保证金率

        股票保证金 = 市值 = 持仓量 * 最新价

        """
        return self._long.margin + self._short.margin

    @property
    def transaction_cost(self):
        """
        [float] 交易费率
        """
        return self._long.transaction_cost + self._short.transaction_cost

    @property
    def positions(self):
        return [self._long, self._short]

    @property
    def long(self):
        return self._long

    @property
    def short(self):
        return self._short


PositionDictType = Dict[str, Dict[POSITION_DIRECTION, Position]]


class PositionProxyDict(UserDict):
    def __init__(self, positions):
        super(PositionProxyDict, self).__init__()
        self._positions = positions  # type: PositionDictType

    def keys(self):
        return self._positions.keys()

    def __getitem__(self, order_book_id):
        position_type, position_proxy_type = self._get_position_types(order_book_id)
        if order_book_id not in self._positions:
            long = position_type(order_book_id, POSITION_DIRECTION.LONG)
            short = position_type(order_book_id, POSITION_DIRECTION.SHORT)
        else:
            positions = self._positions[order_book_id]
            long = positions[POSITION_DIRECTION.LONG]
            short = positions[POSITION_DIRECTION.SHORT]
        return position_proxy_type(long, short)

    def __contains__(self, item):
        return item in self._positions

    def __iter__(self):
        return iter(self._positions)

    def __len__(self):
        return len(self._positions)

    def __setitem__(self, key, value):
        raise TypeError("{} object does not support item assignment".format(self.__class__.__name__))

    def __delitem__(self, key):
        raise TypeError("{} object does not support item deletion".format(self.__class__.__name__))

    def __repr__(self):
        return repr({k: self[k] for k in self._positions.keys()})

    def _get_position_types(self, order_book_id):
        # type: (str) -> Tuple[type, type]
        instrument_type = Environment.get_instance().data_proxy.instrument(order_book_id).type
        position_type = POSITION_TYPE_MAP.get(instrument_type, Position)
        position_proxy_type = POSITION_PROXY_TYPE_MAP.get(instrument_type, PositionProxy)
        return position_type, position_proxy_type



from typing import Optional

from rqalpha.interface import AbstractFrontendValidator
from rqalpha.utils.logger import user_system_log
from rqalpha.model.order import Order
from rqalpha.portfolio.account import Account


class MarginInstrumentValidator(AbstractFrontendValidator):
    """ 融资下单品种限制: 当开启股票池限制且有融资余额时只能交易股票和ETF """

    def validate_submission(self, order: Order, account: Optional[Account] = None) -> Optional[str]:
        if account.cash_liabilities > 0:
            reason = "Order Creation Failed: cash liabilities > 0, {} not support submit order".format(order.order_book_id)
            return reason
        else:
            return None
    
    def validate_cancellation(self, order: Order, account: Optional[Account] = None) -> Optional[str]:
        return None


# -*- coding: utf-8 -*-
# 版权所有 2019 深圳米筐科技有限公司（下称“米筐科技”）
#
# 除非遵守当前许可，否则不得使用本软件。
#
#     * 非商业用途（非商业用途指个人出于非商业目的使用本软件，或者高校、研究所等非营利机构出于教育、科研等目的使用本软件）：
#         遵守 Apache License 2.0（下称“Apache 2.0 许可”），
#         您可以在以下位置获得 Apache 2.0 许可的副本：http://www.apache.org/licenses/LICENSE-2.0。
#         除非法律有要求或以书面形式达成协议，否则本软件分发时需保持当前许可“原样”不变，且不得附加任何条件。
#
#     * 商业用途（商业用途指个人出于任何商业目的使用本软件，或者法人或其他组织出于任何目的使用本软件）：
#         未经米筐科技授权，任何个人不得出于任何商业目的使用本软件（包括但不限于向第三方提供、销售、出租、出借、转让本软件、
#         本软件的衍生产品、引用或借鉴了本软件功能或源代码的产品或服务），任何法人或其他组织不得出于任何目的使用本软件，
#         否则米筐科技有权追究相应的知识产权侵权责任。
#         在此前提下，对本软件的使用同样需要遵守 Apache 2.0 许可，Apache 2.0 许可与本许可冲突之处，以本许可为准。
#         详细的授权流程，请联系 public@ricequant.com 获取。
from __future__ import division

import types
from datetime import date, datetime
from typing import Callable, List, Optional, Union, Iterable

import pandas as pd
import numpy as np
import six

from rqalpha.apis import names
from rqalpha.environment import Environment
from rqalpha.core.execution_context import ExecutionContext
from rqalpha.utils import is_valid_price
from rqalpha.utils.exception import RQInvalidArgument
from rqalpha.utils.i18n import gettext as _
from rqalpha.utils.arg_checker import apply_rules, verify_that
from rqalpha.api import export_as_api
from rqalpha.utils.logger import user_log as logger, user_system_log, user_print
from rqalpha.model.instrument import Instrument
from rqalpha.model.tick import TickObject
from rqalpha.const import (
    EXECUTION_PHASE, ORDER_STATUS, SIDE, POSITION_EFFECT, ORDER_TYPE, MATCHING_TYPE, RUN_TYPE, POSITION_DIRECTION,
    DEFAULT_ACCOUNT_TYPE
)
from rqalpha.model.order import Order, MarketOrder, LimitOrder, OrderStyle, VWAPOrder, TWAPOrder
from rqalpha.core.events import EVENT, Event
from rqalpha.core.strategy_context import StrategyContext
from rqalpha.portfolio.position import Position

export_as_api(logger, name='logger')
export_as_api(user_print, name='print')
export_as_api(LimitOrder, name='LimitOrder')
export_as_api(MarketOrder, name='MarketOrder')
export_as_api(VWAPOrder, name='VWAPOrder')
export_as_api(TWAPOrder, name='TWAPOrder')
export_as_api(ORDER_STATUS, name='ORDER_STATUS')
export_as_api(SIDE, name='SIDE')
export_as_api(POSITION_EFFECT, name='POSITION_EFFECT')
export_as_api(POSITION_DIRECTION, name='POSITION_DIRECTION')
export_as_api(ORDER_TYPE, name='ORDER_TYPE')
export_as_api(RUN_TYPE, name='RUN_TYPE')
export_as_api(MATCHING_TYPE, name='MATCHING_TYPE')
export_as_api(EVENT, name='EVENT')


def assure_instrument(id_or_ins):
    if isinstance(id_or_ins, Instrument):
        return id_or_ins
    elif isinstance(id_or_ins, six.string_types):
        return Environment.get_instance().data_proxy.instrument(id_or_ins)
    else:
        raise RQInvalidArgument(_(u"unsupported order_book_id type"))


def assure_order_book_id(id_or_ins):
    return assure_instrument(id_or_ins).order_book_id


def cal_style(price, style, price_or_style=None):
    if price_or_style is None:
        if price:
            price_or_style = price
        if style:
            price_or_style = style

    if price_or_style is None:
        return MarketOrder()

    if not isinstance(price_or_style, (int, float, OrderStyle)):
        raise RQInvalidArgument(f"price or style or price_or_style type no support. {price_or_style}")

    if isinstance(price_or_style, OrderStyle):
        return price_or_style

    return LimitOrder(price_or_style)


def calc_open_close_style(price, style, price_or_style):
    if isinstance(price_or_style, tuple):
        _length = len(price_or_style)
        if _length == 0:
            o, c = None, None
        elif _length == 1:
            o, c = price_or_style[0], price_or_style[0]
        else:
            o, c = price_or_style[0], price_or_style[1]
        open_style = cal_style(price, style, o)
        close_style = cal_style(price, style, c)
    else:
        open_style = cal_style(price, style, price_or_style)
        close_style = open_style
    return open_style, close_style


@export_as_api
@ExecutionContext.enforce_phase(
    EXECUTION_PHASE.BEFORE_TRADING,
    EXECUTION_PHASE.OPEN_AUCTION,
    EXECUTION_PHASE.ON_BAR,
    EXECUTION_PHASE.ON_TICK,
    EXECUTION_PHASE.AFTER_TRADING,
    EXECUTION_PHASE.SCHEDULED,
)
def get_open_orders():
    # type: () -> List[Order]
    """
    获取当日未成交订单数据
    """
    broker = Environment.get_instance().broker
    return [o for o in broker.get_open_orders() if o.position_effect != POSITION_EFFECT.EXERCISE]


@export_as_api
@apply_rules(
    verify_that("id_or_ins").is_valid_instrument(),
    verify_that("amount").is_number().is_greater_than(0),
    verify_that("side").is_in([SIDE.BUY, SIDE.SELL]),
)
def submit_order(id_or_ins, amount, side, price=None, position_effect=None):
    # type: (Union[str, Instrument], float, SIDE, Optional[float], Optional[POSITION_EFFECT]) -> Optional[Order]
    """
    通用下单函数，策略可以通过该函数自由选择参数下单。

    :param id_or_ins: 下单标的物
    :param amount: 下单量，需为正数
    :param side: 多空方向
    :param price: 下单价格，默认为None，表示市价单
    :param position_effect: 开平方向，交易股票不需要该参数
    :example:

    .. code-block:: python

        # 购买 2000 股的平安银行股票，并以市价单发送：
        submit_order('000001.XSHE', 2000, SIDE.BUY)
        # 平 10 份 RB1812 多方向的今仓，并以 4000 的价格发送限价单
        submit_order('RB1812', 10, SIDE.SELL, price=4000, position_effect=POSITION_EFFECT.CLOSE_TODAY)

    """
    order_book_id = assure_order_book_id(id_or_ins)
    env = Environment.get_instance()
    if (
            env.config.base.run_type != RUN_TYPE.BACKTEST
            and env.get_instrument(order_book_id).type == "Future"
    ):
        if "88" in order_book_id:
            raise RQInvalidArgument(
                _(u"Main Future contracts[88] are not supported in paper trading.")
            )
        if "99" in order_book_id:
            raise RQInvalidArgument(
                _(u"Index Future contracts[99] are not supported in paper trading.")
            )
    style = cal_style(price, None)
    market_price = env.get_last_price(order_book_id)
    if not is_valid_price(market_price):
        reason = _(u"Order Creation Failed: [{order_book_id}] No market data").format(order_book_id=order_book_id)
        env.order_creation_failed(order_book_id, reason)
        return

    amount = int(amount)

    order = Order.__from_create__(
        order_book_id=order_book_id,
        quantity=amount,
        side=side,
        style=style,
        position_effect=position_effect,
    )

    if order.type == ORDER_TYPE.MARKET:
        order.set_frozen_price(market_price)
    if env.can_submit_order(order):
        env.broker.submit_order(order)
        return order


@export_as_api
@ExecutionContext.enforce_phase(
    EXECUTION_PHASE.BEFORE_TRADING,
    EXECUTION_PHASE.OPEN_AUCTION,
    EXECUTION_PHASE.ON_BAR,
    EXECUTION_PHASE.ON_TICK,
    EXECUTION_PHASE.AFTER_TRADING,
    EXECUTION_PHASE.SCHEDULED,
    EXECUTION_PHASE.GLOBAL,
)
@apply_rules(verify_that("order").is_instance_of(Order))
def cancel_order(order):
    # type: (Order) -> Order
    """
    撤单

    :param order: 需要撤销的order对象
    """
    env = Environment.get_instance()
    if env.can_cancel_order(order):
        env.broker.cancel_order(order)
    return order


@export_as_api
@ExecutionContext.enforce_phase(
    EXECUTION_PHASE.ON_INIT,
    EXECUTION_PHASE.BEFORE_TRADING,
    EXECUTION_PHASE.OPEN_AUCTION,
    EXECUTION_PHASE.OPEN_AUCTION,
    EXECUTION_PHASE.ON_BAR,
    EXECUTION_PHASE.ON_TICK,
    EXECUTION_PHASE.AFTER_TRADING,
    EXECUTION_PHASE.SCHEDULED,
)
@apply_rules(verify_that("id_or_symbols").are_valid_instruments())
def update_universe(id_or_symbols):
    # type: (Union[str, Instrument, Iterable[str], Iterable[Instrument]]) -> None
    """
    该方法用于更新现在关注的证券的集合（e.g.：股票池）。PS：会在下一个bar事件触发时候产生（新的关注的股票池更新）效果。并且update_universe会是覆盖（overwrite）的操作而不是在已有的股票池的基础上进行增量添加。比如已有的股票池为['000001.XSHE', '000024.XSHE']然后调用了update_universe(['000030.XSHE'])之后，股票池就会变成000030.XSHE一个股票了，随后的数据更新也只会跟踪000030.XSHE这一个股票了。

    :param id_or_symbols: 标的物

    """
    if isinstance(id_or_symbols, (six.string_types, Instrument)):
        id_or_symbols = [id_or_symbols]
    order_book_ids = set(
        assure_order_book_id(order_book_id) for order_book_id in id_or_symbols
    )
    if order_book_ids != Environment.get_instance().get_universe():
        Environment.get_instance().update_universe(order_book_ids)


@export_as_api
@ExecutionContext.enforce_phase(
    EXECUTION_PHASE.ON_INIT,
    EXECUTION_PHASE.BEFORE_TRADING,
    EXECUTION_PHASE.OPEN_AUCTION,
    EXECUTION_PHASE.ON_BAR,
    EXECUTION_PHASE.ON_TICK,
    EXECUTION_PHASE.AFTER_TRADING,
    EXECUTION_PHASE.SCHEDULED,
)
@apply_rules(verify_that("id_or_symbols").are_valid_instruments())
def subscribe(id_or_symbols):
    # type: (Union[str, Instrument, Iterable[str], Iterable[Instrument]]) -> None

    """
    订阅合约行情。

    在日级别回测中不需要订阅合约。

    在分钟回测中，若策略只设置了股票账户则不需要订阅合约；若设置了期货账户，则需要订阅策略关注的期货合约，框架会根据订阅的期货合约品种触发对应交易时间的 handle_bar。为了方便起见，也可以以直接订阅主力连续合约。

    在 tick 回测中，策略需要订阅每一个关注的股票/期货合约，框架会根据订阅池触发对应标的的 handle_tick。

    :param id_or_symbols: 标的物

    """
    current_universe = Environment.get_instance().get_universe()
    if isinstance(id_or_symbols, six.string_types):
        order_book_id = instruments(id_or_symbols).order_book_id
        current_universe.add(order_book_id)
    elif isinstance(id_or_symbols, Instrument):
        current_universe.add(id_or_symbols.order_book_id)
    elif isinstance(id_or_symbols, Iterable):
        for item in id_or_symbols:
            current_universe.add(assure_order_book_id(item))
    else:
        raise RQInvalidArgument(_(u"unsupported order_book_id type"))
    verify_that("id_or_symbols")._are_valid_instruments("subscribe", id_or_symbols)
    Environment.get_instance().update_universe(current_universe)


@export_as_api
@ExecutionContext.enforce_phase(
    EXECUTION_PHASE.ON_INIT,
    EXECUTION_PHASE.BEFORE_TRADING,
    EXECUTION_PHASE.OPEN_AUCTION,
    EXECUTION_PHASE.ON_BAR,
    EXECUTION_PHASE.ON_TICK,
    EXECUTION_PHASE.AFTER_TRADING,
    EXECUTION_PHASE.SCHEDULED,
)
@apply_rules(verify_that("id_or_symbols").are_valid_instruments())
def unsubscribe(id_or_symbols):
    # type: (Union[str, Instrument, Iterable[str], Iterable[Instrument]]) -> None

    """
    取消订阅合约行情。取消订阅会导致合约池内合约的减少，如果当前合约池中没有任何合约，则策略直接退出。

    :param id_or_symbols: 标的物

    """
    current_universe = Environment.get_instance().get_universe()
    if isinstance(id_or_symbols, six.string_types):
        order_book_id = instruments(id_or_symbols).order_book_id
        current_universe.discard(order_book_id)
    elif isinstance(id_or_symbols, Instrument):
        current_universe.discard(id_or_symbols.order_book_id)
    elif isinstance(id_or_symbols, Iterable):
        for item in id_or_symbols:
            i = assure_order_book_id(item)
            current_universe.discard(i)
    else:
        raise RQInvalidArgument(_(u"unsupported order_book_id type"))

    Environment.get_instance().update_universe(current_universe)


@export_as_api
@ExecutionContext.enforce_phase(
    EXECUTION_PHASE.ON_INIT,
    EXECUTION_PHASE.BEFORE_TRADING,
    EXECUTION_PHASE.OPEN_AUCTION,
    EXECUTION_PHASE.ON_BAR,
    EXECUTION_PHASE.ON_TICK,
    EXECUTION_PHASE.AFTER_TRADING,
    EXECUTION_PHASE.SCHEDULED,
)
@apply_rules(
    verify_that("date").is_valid_date(ignore_none=True),
    verify_that("tenor").is_in(names.VALID_TENORS, ignore_none=True),
)
def get_yield_curve(date=None, tenor=None):
    # type: (Optional[Union[str, date, datetime, pd.Timestamp]], str) -> pd.DataFrame
    """
    获取某个国家市场指定日期的收益率曲线水平。

    数据为2002年至今的中债国债收益率曲线，来源于中央国债登记结算有限责任公司。

    :param date: 查询日期，默认为策略当前日期前一天
    :param tenor: 标准期限，'0S' - 隔夜，'1M' - 1个月，'1Y' - 1年，默认为全部期限

    :example:

    ..  code-block:: python3
        :linenos:

        [In]
        get_yield_curve('20130104')

        [Out]
                        0S      1M      2M      3M      6M      9M      1Y      2Y  \
        2013-01-04  0.0196  0.0253  0.0288  0.0279  0.0280  0.0283  0.0292  0.0310

                        3Y      4Y   ...        6Y      7Y      8Y      9Y     10Y  \
        2013-01-04  0.0314  0.0318   ...    0.0342  0.0350  0.0353  0.0357  0.0361
        ...
    """
    env = Environment.get_instance()
    trading_date = env.trading_dt.date()
    yesterday = env.data_proxy.get_previous_trading_date(trading_date)

    if date is None:
        date = yesterday
    else:
        date = pd.Timestamp(date)
        if date > yesterday:
            raise RQInvalidArgument(
                "get_yield_curve: {} >= now({})".format(date, yesterday)
            )

    return env.data_proxy.get_yield_curve(start_date=date, end_date=date, tenor=tenor)


@export_as_api
@ExecutionContext.enforce_phase(
    EXECUTION_PHASE.BEFORE_TRADING,
    EXECUTION_PHASE.OPEN_AUCTION,
    EXECUTION_PHASE.ON_BAR,
    EXECUTION_PHASE.ON_TICK,
    EXECUTION_PHASE.AFTER_TRADING,
    EXECUTION_PHASE.SCHEDULED,
)
@apply_rules(
    verify_that("order_book_id", pre_check=True).is_listed_instrument(),
    verify_that("bar_count").is_instance_of(int).is_greater_than(0),
    verify_that("frequency", pre_check=True).is_valid_frequency(),
    verify_that("fields").are_valid_fields(
        names.VALID_HISTORY_FIELDS, ignore_none=True
    ),
    verify_that("skip_suspended").is_instance_of(bool),
    verify_that("include_now").is_instance_of(bool),
    verify_that("adjust_type").is_in({"pre", "none", "post"}),
)
def history_bars(
        order_book_id,
        bar_count,
        frequency,
        fields=None,
        skip_suspended=True,
        include_now=False,
        adjust_type="pre",
):
    # type:(str, int, str, Optional[Union[str, List[str]]], Optional[bool], Optional[bool], Optional[str]) -> np.ndarray
    """
    获取指定合约的历史 k 线行情，同时支持日以及分钟历史数据。不能在init中调用。

    日回测获取分钟历史数据：不支持

    日回测获取日历史数据

    =========================   ===================================================
    调用时间                      返回数据
    =========================   ===================================================
    T日before_trading            T-1日day bar
    T日handle_bar                T日day bar
    =========================   ===================================================

    分钟回测获取日历史数据

    =========================   ===================================================
    调用时间                      返回数据
    =========================   ===================================================
    T日before_trading            T-1日day bar
    T日handle_bar                T-1日day bar
    =========================   ===================================================

    分钟回测获取分钟历史数据

    =========================   ===================================================
    调用时间                      返回数据
    =========================   ===================================================
    T日before_trading            T-1日最后一个minute bar
    T日handle_bar                T日当前minute bar
    =========================   ===================================================

    :param order_book_id: 合约代码
    :param bar_count: 获取的历史数据数量，必填项
    :param frequency: 获取数据什么样的频率进行。'1d'、'1m' 和 '1w' 分别表示每日、每分钟和每周，必填项
    :param fields: 返回数据字段。必填项。见下方列表。
    :param skip_suspended: 是否跳过停牌数据
    :param include_now: 是否包含当前数据
    :param adjust_type: 复权类型，默认为前复权 pre；可选 pre, none, post

    =========================   ===================================================
    fields                      字段名
    =========================   ===================================================
    datetime                    时间戳
    open                        开盘价
    high                        最高价
    low                         最低价
    close                       收盘价
    volume                      成交量
    total_turnover              成交额
    open_interest               持仓量（期货专用）
    basis_spread                期现差（股指期货专用）
    settlement                  结算价（期货日线专用）
    prev_settlement             结算价（期货日线专用）
    =========================   ===================================================

    :example:

    获取最近5天的日线收盘价序列（策略当前日期为20160706）:

    ..  code-block:: python3
        :linenos:

        [In]
        logger.info(history_bars('000002.XSHE', 5, '1d', 'close'))
        [Out]
        [ 8.69  8.7   8.71  8.81  8.81]
    """
    order_book_id = assure_order_book_id(order_book_id)
    env = Environment.get_instance()
    dt = env.calendar_dt

    if frequency[-1] == "m" and env.config.base.frequency == "1d":
        raise RQInvalidArgument("can not get minute history in day back test")

    if frequency[-1] == "d" and frequency != "1d":
        raise RQInvalidArgument("invalid frequency")

    if adjust_type not in {"pre", "post", "none"}:
        raise RuntimeError("invalid adjust_type")

    if frequency == "1d":
        sys_frequency = Environment.get_instance().config.base.frequency
        if (
                sys_frequency in ["1m", "tick"]
                and not include_now
                and ExecutionContext.phase() != EXECUTION_PHASE.AFTER_TRADING
        ) or (ExecutionContext.phase() in (EXECUTION_PHASE.BEFORE_TRADING, EXECUTION_PHASE.OPEN_AUCTION)):
            dt = env.data_proxy.get_previous_trading_date(env.trading_dt.date())
            # 当 EXECUTION_PHASE.BEFORE_TRADING 的时候，强制 include_now 为 False
            include_now = False
        if sys_frequency == "1d":
            # 日回测不支持 include_now
            include_now = False

    if fields is None:
        fields = ["datetime", "open", "high", "low", "close", "volume"]

    return env.data_proxy.history_bars(
        order_book_id,
        bar_count,
        frequency,
        fields,
        dt,
        skip_suspended=skip_suspended,
        include_now=include_now,
        adjust_type=adjust_type,
        adjust_orig=env.trading_dt,
    )


@export_as_api
@apply_rules(
    verify_that("order_book_id", pre_check=True).is_listed_instrument(),
    verify_that('count').is_instance_of(int).is_greater_than(0)
)
def history_ticks(order_book_id, count):
    # type: (str, int) -> List[TickObject]
    """
    获取指定合约历史（不晚于当前时间的）tick 对象，仅支持在 tick 级别的策略（回测、模拟交易、实盘）中调用。

    :param order_book_id: 合约代码
    :param count: 获取的 tick 数量

    """
    env = Environment.get_instance()
    sys_frequency = env.config.base.frequency
    if sys_frequency == "1d":
        raise RuntimeError("history_ticks does not support day bar backtest.")

    order_book_id = assure_order_book_id(order_book_id)
    dt = env.calendar_dt

    return env.data_proxy.history_ticks(order_book_id, count, dt)


@export_as_api
@ExecutionContext.enforce_phase(
    EXECUTION_PHASE.ON_INIT,
    EXECUTION_PHASE.BEFORE_TRADING,
    EXECUTION_PHASE.OPEN_AUCTION,
    EXECUTION_PHASE.ON_BAR,
    EXECUTION_PHASE.ON_TICK,
    EXECUTION_PHASE.AFTER_TRADING,
    EXECUTION_PHASE.SCHEDULED,
)
@apply_rules(
    verify_that("type").are_valid_fields(
        names.VALID_INSTRUMENT_TYPES, ignore_none=True
    ),
    verify_that("date").is_valid_date(ignore_none=True),
)
def all_instruments(type=None, date=None):
    # type: (str, Union[str, datetime, date]) -> pd.DataFrame
    """
    获取某个国家市场的所有合约信息。使用者可以通过这一方法很快地对合约信息有一个快速了解，目前仅支持中国市场。

    :param type: 需要查询合约类型，例如：type='CS'代表股票。默认是所有类型
    :param date: 查询时间点

    其中type参数传入的合约类型和对应的解释如下：

    =========================   ===================================================
    合约类型                      说明
    =========================   ===================================================
    CS                          Common Stock, 即股票
    ETF                         Exchange Traded Fund, 即交易所交易基金
    LOF                         Listed Open-Ended Fund，即上市型开放式基金
    INDX                        Index, 即指数
    Future                      Futures，即期货，包含股指、国债和商品期货
    =========================   ===================================================

    """
    env = Environment.get_instance()
    if date is None:
        dt = env.trading_dt
    else:
        dt = pd.Timestamp(date).to_pydatetime()
        dt = min(dt, env.trading_dt)

    if type is not None:
        if isinstance(type, six.string_types):
            type = [type]

        types = set()
        for t in type:
            if t == "Stock":
                types.add("CS")
            elif t == "Fund":
                types.update(["ETF", "LOF"])
            else:
                types.add(t)
    else:
        types = None

    result = env.data_proxy.all_instruments(types, dt)
    if types is not None and len(types) == 1:
        data = []
        for i in result:
            instrument_dic = {k: v for k, v in i.__dict__.items() if not k.startswith("_")}
            data.append(instrument_dic)
        return pd.DataFrame(data)

    return pd.DataFrame(
        [
            [i.order_book_id, i.symbol, i.type, i.listed_date, i.de_listed_date]
            for i in result
        ],
        columns=["order_book_id", "symbol", "type", "listed_date", "de_listed_date"],
    )


@export_as_api
@ExecutionContext.enforce_phase(
    EXECUTION_PHASE.ON_INIT,
    EXECUTION_PHASE.BEFORE_TRADING,
    EXECUTION_PHASE.OPEN_AUCTION,
    EXECUTION_PHASE.ON_BAR,
    EXECUTION_PHASE.ON_TICK,
    EXECUTION_PHASE.AFTER_TRADING,
    EXECUTION_PHASE.SCHEDULED,
)
@apply_rules(verify_that("id_or_symbols").is_instance_of((str, Iterable)))
def instruments(id_or_symbols):
    # type: (Union[str, List[str]]) -> Union[Instrument, List[Instrument]]
    """
    获取某个国家市场内一个或多个合约的详细信息。目前仅支持中国市场。

    :param id_or_symbols: 合约代码或者合约代码列表

    :example:

    *   获取单一股票合约的详细信息:

        ..  code-block:: python3
            :linenos:

            [In]instruments('000001.XSHE')
            [Out]
            Instrument(order_book_id=000001.XSHE, symbol=平安银行, abbrev_symbol=PAYH, listed_date=19910403, de_listed_date=null, board_type=MainBoard, sector_code_name=金融, sector_code=Financials, round_lot=100, exchange=XSHE, special_type=Normal, status=Active)

    *   获取多个股票合约的详细信息:

        ..  code-block:: python3
            :linenos:

            [In]instruments(['000001.XSHE', '000024.XSHE'])
            [Out]
            [Instrument(order_book_id=000001.XSHE, symbol=平安银行, abbrev_symbol=PAYH, listed_date=19910403, de_listed_date=null, board_type=MainBoard, sector_code_name=金融, sector_code=Financials, round_lot=100, exchange=XSHE, special_type=Normal, status=Active), Instrument(order_book_id=000024.XSHE, symbol=招商地产, abbrev_symbol=ZSDC, listed_date=19930607, de_listed_date=null, board_type=MainBoard, sector_code_name=金融, sector_code=Financials, round_lot=100, exchange=XSHE, special_type=Normal, status=Active)]

    *   获取合约已上市天数:

        ..  code-block:: python
            :linenos:

            instruments('000001.XSHE').days_from_listed()

    *   获取合约距离到期天数:

        ..  code-block:: python
            :linenos:

            instruments('IF1701').days_to_expire()
    """
    return Environment.get_instance().data_proxy.instruments(id_or_symbols)


@export_as_api
@apply_rules(
    verify_that("start_date").is_valid_date(ignore_none=False),
    verify_that("end_date").is_valid_date(ignore_none=False),
)
def get_trading_dates(start_date, end_date):
    # type: (Union[str, date, datetime, pd.Timestamp], Union[str, date, datetime, pd.Timestamp]) -> pd.DatetimeIndex
    """
    获取某个国家市场的交易日列表（起止日期加入判断）。目前仅支持中国市场。

    :param start_date: 开始日期
    :param end_date: 结束如期

    """
    return Environment.get_instance().data_proxy.get_trading_dates(start_date, end_date)


@export_as_api
@apply_rules(
    verify_that("date").is_valid_date(ignore_none=False),
    verify_that("n").is_instance_of(int).is_greater_or_equal_than(1),
)
def get_previous_trading_date(date, n=1):
    # type: (Union[str, date, datetime, pd.Timestamp], Optional[int]) -> date
    """
    获取指定日期的之前的第 n 个交易日。

    :param date: 指定日期
    :param n:

    :example:

    ..  code-block:: python3
        :linenos:

        [In]get_previous_trading_date(date='2016-05-02')
        [Out]
        [datetime.date(2016, 4, 29)]
    """
    return Environment.get_instance().data_proxy.get_previous_trading_date(date, n)


@export_as_api
@apply_rules(
    verify_that("date").is_valid_date(ignore_none=False),
    verify_that("n").is_instance_of(int).is_greater_or_equal_than(1),
)
def get_next_trading_date(date, n=1):
    # type: (Union[str, date, datetime, pd.Timestamp], Optional[int]) -> date
    """
    获取指定日期之后的第 n 个交易日

    :param date: 指定日期
    :param n:

    :example:

    ..  code-block:: python3
        :linenos:

        [In]get_next_trading_date(date='2016-05-01')
        [Out]
        [datetime.date(2016, 5, 3)]
    """
    return Environment.get_instance().data_proxy.get_next_trading_date(date, n)


@export_as_api
@ExecutionContext.enforce_phase(
    EXECUTION_PHASE.BEFORE_TRADING,
    EXECUTION_PHASE.OPEN_AUCTION,
    EXECUTION_PHASE.ON_BAR,
    EXECUTION_PHASE.ON_TICK,
    EXECUTION_PHASE.AFTER_TRADING,
    EXECUTION_PHASE.SCHEDULED,
)
@apply_rules(verify_that("id_or_symbol").is_valid_instrument())
def current_snapshot(id_or_symbol):
    # type: (Union[str, Instrument]) -> Optional[TickObject]
    """
    获得当前市场快照数据。只能在日内交易阶段调用，获取当日调用时点的市场快照数据。
    市场快照数据记录了每日从开盘到当前的数据信息，可以理解为一个动态的day bar数据。
    在目前分钟回测中，快照数据为当日所有分钟线累积而成，一般情况下，最后一个分钟线获取到的快照数据应当与当日的日线行情保持一致。
    需要注意，在实盘模拟中，该函数返回的是调用当时的市场快照情况，所以在同一个handle_bar中不同时点调用可能返回的数据不同。
    如果当日截止到调用时候对应股票没有任何成交，那么snapshot中的close, high, low, last几个价格水平都将以0表示。

    :param d_or_symbol: 合约代码或简称

    :example:

    在handle_bar中调用该函数，假设策略当前时间是20160104 09:33:

    ..  code-block:: python3
        :linenos:

        [In]
        logger.info(current_snapshot('000001.XSHE'))
        [Out]
        2016-01-04 09:33:00.00  INFO
        Snapshot(order_book_id: '000001.XSHE', datetime: datetime.datetime(2016, 1, 4, 9, 33), open: 10.0, high: 10.025, low: 9.9667, last: 9.9917, volume: 2050320, total_turnover: 20485195, prev_close: 9.99)
    """
    env = Environment.get_instance()
    frequency = env.config.base.frequency
    order_book_id = assure_order_book_id(id_or_symbol)

    dt = env.calendar_dt

    if env.config.base.run_type == RUN_TYPE.BACKTEST:
        if ExecutionContext.phase() == EXECUTION_PHASE.BEFORE_TRADING:
            dt = env.data_proxy.get_previous_trading_date(env.trading_dt.date())
            return env.data_proxy.current_snapshot(order_book_id, "1d", dt)
        elif ExecutionContext.phase() == EXECUTION_PHASE.AFTER_TRADING:
            return env.data_proxy.current_snapshot(order_book_id, "1d", dt)

    # PT、实盘直接取最新快照，忽略 frequency, dt 参数
    return env.data_proxy.current_snapshot(order_book_id, frequency, dt)


@export_as_api
def get_positions():
    # type: () -> List[Position]
    """
    获取所有持仓对象列表。

    :example:

    ..  code-block:: python3

        [In] get_positions()
        [Out]
        [StockPosition(order_book_id=000001.XSHE, direction=LONG, quantity=1000, market_value=19520.0, trading_pnl=0.0, position_pnl=0),
        StockPosition(order_book_id=RB2112, direction=SHORT, quantity=2, market_value=-111580.0, trading_pnl=0.0, position_pnl=0)]

    """
    portfolio = Environment.get_instance().portfolio
    return portfolio.get_positions()


@export_as_api
@apply_rules(
    verify_that("direction").is_in([POSITION_DIRECTION.LONG, POSITION_DIRECTION.SHORT])
)
def get_position(order_book_id, direction=POSITION_DIRECTION.LONG):
    # type: (str, Optional[POSITION_DIRECTION]) -> Position
    """
    获取某个标的的持仓对象。

    :param order_book_id: 标的编号
    :param direction: 持仓方向

    :example:

    ..  code-block:: python3

        [In] get_position('000001.XSHE', POSITION_DIRECTION.LONG)
        [Out]
        StockPosition(order_book_id=000001.XSHE, direction=LONG, quantity=268600, market_value=4995960.0, trading_pnl=0.0, position_pnl=0)

    """
    portfolio = Environment.get_instance().portfolio
    return portfolio.get_position(order_book_id, direction)


@export_as_api
@ExecutionContext.enforce_phase(
    EXECUTION_PHASE.ON_INIT
)
@apply_rules(
    verify_that("event_type").is_instance_of(EVENT),
    verify_that("handler").is_instance_of(types.FunctionType),
)
def subscribe_event(event_type, handler):
    # type: (EVENT, Callable[[StrategyContext, Event], None]) -> None
    """
    订阅框架内部事件，注册事件处理函数

    :param event_type: 事件类型
    :param handler: 处理函数

    """
    env = Environment.get_instance()
    user_strategy = env.user_strategy
    env.event_bus.add_listener(
        event_type, user_strategy.wrap_user_event_handler(handler), user=True
    )


@export_as_api
def symbol(order_book_id, sep=", "):
    if isinstance(order_book_id, six.string_types):
        return "{}[{}]".format(order_book_id, Environment.get_instance().get_instrument(order_book_id).symbol)
    else:
        s = sep.join(symbol(item) for item in order_book_id)
        return s


@export_as_api
@ExecutionContext.enforce_phase(
    EXECUTION_PHASE.OPEN_AUCTION,
    EXECUTION_PHASE.ON_BAR,
    EXECUTION_PHASE.ON_TICK,
    EXECUTION_PHASE.SCHEDULED,
    EXECUTION_PHASE.GLOBAL
)
def deposit(account_type: str, amount: float, receiving_days: int = 0):
    """
    入金（增加账户资金）

    :param account_type: 账户类型
    :param amount: 入金金额
    :param receiving_days: 入金到账天数，0 表示立刻到账，1 表示资金在下一个交易日盘前到账
    :return: None
    """
    env = Environment.get_instance()
    return env.portfolio.deposit_withdraw(account_type, amount, receiving_days)


@export_as_api
@ExecutionContext.enforce_phase(
    EXECUTION_PHASE.OPEN_AUCTION,
    EXECUTION_PHASE.ON_BAR,
    EXECUTION_PHASE.ON_TICK,
    EXECUTION_PHASE.SCHEDULED,
    EXECUTION_PHASE.GLOBAL
)
@apply_rules(
    verify_that("account_type").is_in(DEFAULT_ACCOUNT_TYPE),
    verify_that("amount").is_number(),
)
def withdraw(account_type, amount):
    # type: (str, float) -> None
    """
    出金（减少账户资金）

    :param account_type: 账户类型
    :param amount: 减少金额
    :return: None
    """
    env = Environment.get_instance()
    return env.portfolio.deposit_withdraw(account_type, amount * -1)


@export_as_api
@ExecutionContext.enforce_phase(
    EXECUTION_PHASE.OPEN_AUCTION,
    EXECUTION_PHASE.ON_BAR,
    EXECUTION_PHASE.ON_TICK,
    EXECUTION_PHASE.SCHEDULED,
    EXECUTION_PHASE.GLOBAL
)
@apply_rules(
    verify_that("account_type").is_in(DEFAULT_ACCOUNT_TYPE),
    verify_that("amount", pre_check=True).is_instance_of((int, float)).is_greater_than(0),
)
def finance(amount, account_type=DEFAULT_ACCOUNT_TYPE.STOCK):
    """
    融资

    :param amount: 融资金额
    :param account_type: 融资账户
    :return: None
    """
    env = Environment.get_instance()
    return env.portfolio.finance_repay(amount, account_type)


@export_as_api
@ExecutionContext.enforce_phase(
    EXECUTION_PHASE.OPEN_AUCTION,
    EXECUTION_PHASE.ON_BAR,
    EXECUTION_PHASE.ON_TICK,
    EXECUTION_PHASE.SCHEDULED,
)
@apply_rules(
    verify_that("account_type").is_in(DEFAULT_ACCOUNT_TYPE),
    verify_that("amount", pre_check=True).is_instance_of((int, float)).is_greater_than(0),
)
def repay(amount, account_type=DEFAULT_ACCOUNT_TYPE.STOCK):
    """
    还款

    :param amount: 还款金额
    :param account_type: 还款账户
    :return: None
    """
    env = Environment.get_instance()
    return env.portfolio.finance_repay(amount * -1, account_type)





# -*- coding: utf-8 -*-
# 版权所有 2019 深圳米筐科技有限公司（下称“米筐科技”）
#
# 除非遵守当前许可，否则不得使用本软件。
#
#     * 非商业用途（非商业用途指个人出于非商业目的使用本软件，或者高校、研究所等非营利机构出于教育、科研等目的使用本软件）：
#         遵守 Apache License 2.0（下称“Apache 2.0 许可”），您可以在以下位置获得 Apache 2.0 许可的副本：http://www.apache.org/licenses/LICENSE-2.0。
#         除非法律有要求或以书面形式达成协议，否则本软件分发时需保持当前许可“原样”不变，且不得附加任何条件。
#
#     * 商业用途（商业用途指个人出于任何商业目的使用本软件，或者法人或其他组织出于任何目的使用本软件）：
#         未经米筐科技授权，任何个人不得出于任何商业目的使用本软件（包括但不限于向第三方提供、销售、出租、出借、转让本软件、本软件的衍生产品、引用或借鉴了本软件功能或源代码的产品或服务），任何法人或其他组织不得出于任何目的使用本软件，否则米筐科技有权追究相应的知识产权侵权责任。
#         在此前提下，对本软件的使用同样需要遵守 Apache 2.0 许可，Apache 2.0 许可与本许可冲突之处，以本许可为准。
#         详细的授权流程，请联系 public@ricequant.com 获取。


import six
from rqalpha.core.events import EVENT
from rqalpha.utils.logger import user_system_log

from rqalpha.interface import AbstractMod
from rqalpha.utils.i18n import gettext as _
from rqalpha.utils.exception import patch_user_exc
from rqalpha.const import MATCHING_TYPE, RUN_TYPE

from rqalpha.mod.rqalpha_mod_sys_simulation.simulation_broker import SimulationBroker
from rqalpha.mod.rqalpha_mod_sys_simulation.signal_broker import SignalBroker
from rqalpha.mod.rqalpha_mod_sys_simulation.simulation_event_source import SimulationEventSource


class SimulationMod(AbstractMod):
    def __init__(self):
        self._env = None

    def start_up(self, env, mod_config):
        self._env = env
        if env.config.base.run_type == RUN_TYPE.LIVE_TRADING:
            return

        mod_config.matching_type = self.parse_matching_type(mod_config.matching_type, env.config.base.frequency)

        if env.config.base.margin_multiplier <= 0:
            raise patch_user_exc(ValueError(_(u"invalid margin multiplier value: value range is (0, +∞]")))

        if env.config.base.frequency == "tick":
            if mod_config.matching_type not in [
                MATCHING_TYPE.NEXT_TICK_LAST,
                MATCHING_TYPE.NEXT_TICK_BEST_OWN,
                MATCHING_TYPE.NEXT_TICK_BEST_COUNTERPARTY,
                MATCHING_TYPE.COUNTERPARTY_OFFER,
            ]:
                raise RuntimeError(_("Not supported matching type {}").format(mod_config.matching_type))
        else:
            if mod_config.matching_type not in [
                MATCHING_TYPE.NEXT_BAR_OPEN,
                MATCHING_TYPE.VWAP,
                MATCHING_TYPE.CURRENT_BAR_CLOSE,
            ]:
                raise RuntimeError(_("Not supported matching type {}").format(mod_config.matching_type))

        if env.config.base.frequency == "1d" and mod_config.matching_type == MATCHING_TYPE.NEXT_BAR_OPEN:
            mod_config.matching_type = MATCHING_TYPE.CURRENT_BAR_CLOSE
            user_system_log.warn(_(u"matching_type = 'next_bar' is abandoned when frequency == '1d',"
                                   u"Current matching_type is 'current_bar'."))

        if mod_config.signal:
            env.set_broker(SignalBroker(env, mod_config))
        else:
            env.set_broker(SimulationBroker(env, mod_config))

        if mod_config.management_fee:
            env.event_bus.add_listener(EVENT.POST_SYSTEM_INIT, self.register_management_fee_calculator)

        event_source = SimulationEventSource(env)
        env.set_event_source(event_source)

        from .validator import OrderStyleValidator
        env.add_frontend_validator(OrderStyleValidator(env.config.base.frequency))

    def tear_down(self, code, exception=None):
        pass

    @staticmethod
    def parse_matching_type(me_str, frequency):
        if me_str is None:
            # None 表示根据 frequency 自动选择
            if frequency in ["1d", "1m"]:
                me_str = "current_bar"
            elif frequency == "tick":
                me_str = "last"
            else:
                raise ValueError("frequency only support ['1d', '1m', 'tick']")

        assert isinstance(me_str, six.string_types)
        me_str = me_str.lower()
        if me_str == "current_bar":
            return MATCHING_TYPE.CURRENT_BAR_CLOSE
        if me_str == "vwap":
            return MATCHING_TYPE.VWAP
        elif me_str == "next_bar":
            return MATCHING_TYPE.NEXT_BAR_OPEN
        elif me_str == "last":
            return MATCHING_TYPE.NEXT_TICK_LAST
        elif me_str == "best_own":
            return MATCHING_TYPE.NEXT_TICK_BEST_OWN
        elif me_str == "best_counterparty":
            return MATCHING_TYPE.NEXT_TICK_BEST_COUNTERPARTY
        elif me_str == "counterparty_offer":
            return MATCHING_TYPE.COUNTERPARTY_OFFER
        else:
            raise NotImplementedError

    def register_management_fee_calculator(self, event):
        management_fee = self._env.config.mod.sys_simulation.management_fee
        accounts = self._env.portfolio.accounts
        for _account_type, v in management_fee:
            _account_type = _account_type.upper()
            if _account_type not in accounts:
                all_account_type = list(accounts.keys())
                raise ValueError(_("NO account_type = ({}) in {}").format(_account_type, all_account_type))
            accounts[_account_type].set_management_fee_rate(float(v))



# -*- coding: utf-8 -*-
# 版权所有 2019 深圳米筐科技有限公司（下称“米筐科技”）
#
# 除非遵守当前许可，否则不得使用本软件。
#
#     * 非商业用途（非商业用途指个人出于非商业目的使用本软件，或者高校、研究所等非营利机构出于教育、科研等目的使用本软件）：
#         遵守 Apache License 2.0（下称“Apache 2.0 许可”），
#         您可以在以下位置获得 Apache 2.0 许可的副本：http://www.apache.org/licenses/LICENSE-2.0。
#         除非法律有要求或以书面形式达成协议，否则本软件分发时需保持当前许可“原样”不变，且不得附加任何条件。
#
#     * 商业用途（商业用途指个人出于任何商业目的使用本软件，或者法人或其他组织出于任何目的使用本软件）：
#         未经米筐科技授权，任何个人不得出于任何商业目的使用本软件（包括但不限于向第三方提供、销售、出租、出借、转让本软件、
#         本软件的衍生产品、引用或借鉴了本软件功能或源代码的产品或服务），任何法人或其他组织不得出于任何目的使用本软件，
#         否则米筐科技有权追究相应的知识产权侵权责任。
#         在此前提下，对本软件的使用同样需要遵守 Apache 2.0 许可，Apache 2.0 许可与本许可冲突之处，以本许可为准。
#         详细的授权流程，请联系 public@ricequant.com 获取。

from datetime import date

from numpy.testing import assert_almost_equal

from rqalpha.environment import Environment

from ..utils import make_test_strategy_decorator

test_strategies = []

__config__ = {
    "base": {
        "start_date": "2018-04-01",
        "end_date": "2018-05-01",
        "frequency": "1d",
        "accounts": {
            "future": 10000000000
        }
    },
    "extra": {
        "log_level": "error",
    },
    "mod": {
        "sys_progress": {
            "enabled": True,
            "show": True,
        },
    },
}


def assert_almost_equal(first, second):
    assert round(abs(first - second), 10) == 0


def test_futures_info():
    __config__ = {
        "base": {
            "future_info": {
                "SC": {
                    "close_commission_ratio": 0.0001,
                    "open_commission_ratio": 0.0002,
                    "commission_type": "BY_MONEY",
                }
            }
        }
        }

    def init(context):
        context.f1 = "SC1809"
        subscribe_event(EVENT.TRADE, on_trade)

    def handle_bar(*_):
        buy_open("SC1809", 2)
        sell_close("SC1809", 2, close_today=False)

    def on_trade(_, event):
        trade = event.trade
        contract_multiplier = Environment.get_instance().data_proxy.instrument("SC1809").contract_multiplier
        if trade.position_effect == POSITION_EFFECT.OPEN:
            assert_almost_equal(
                trade.transaction_cost, 0.0002 * trade.last_quantity * trade.last_price * contract_multiplier
            )
        elif trade.position_effect == POSITION_EFFECT.CLOSE:
            assert_almost_equal(
                trade.transaction_cost, 0.0001 * trade.last_quantity * trade.last_price * contract_multiplier
            )
        else:
            assert trade.transaction_cost == 0

    return locals()


def test_init_position():
    __config__ = {
        "base": {
            "accounts": {
                "stock": 10000000
            },
            "init_positions": "000006.XSHE:10000"
        },
   }

    def before_trading(context):
        if context.now.date() == date(2018, 4, 2):
            # 首个交易日，持仓价格应为昨收
            pos = get_position("000006.XSHE")  # noqa
            assert_almost_equal(pos.last_price, 7.37)
            assert pos.quantity == 10000

    def init(context):
        context.f1 = "000006.XSHE"
        stock_position = context.portfolio.get_position('000006.XSHE', 'LONG')
        assert stock_position.quantity == 10000

    return locals()



# -*- coding: utf-8 -*-
# 版权所有 2019 深圳米筐科技有限公司（下称“米筐科技”）
#
# 除非遵守当前许可，否则不得使用本软件。
#
#     * 非商业用途（非商业用途指个人出于非商业目的使用本软件，或者高校、研究所等非营利机构出于教育、科研等目的使用本软件）：
#         遵守 Apache License 2.0（下称“Apache 2.0 许可”），
#         您可以在以下位置获得 Apache 2.0 许可的副本：http://www.apache.org/licenses/LICENSE-2.0。
#         除非法律有要求或以书面形式达成协议，否则本软件分发时需保持当前许可“原样”不变，且不得附加任何条件。
#
#     * 商业用途（商业用途指个人出于任何商业目的使用本软件，或者法人或其他组织出于任何目的使用本软件）：
#         未经米筐科技授权，任何个人不得出于任何商业目的使用本软件（包括但不限于向第三方提供、销售、出租、出借、转让本软件、
#         本软件的衍生产品、引用或借鉴了本软件功能或源代码的产品或服务），任何法人或其他组织不得出于任何目的使用本软件，
#         否则米筐科技有权追究相应的知识产权侵权责任。
#         在此前提下，对本软件的使用同样需要遵守 Apache 2.0 许可，Apache 2.0 许可与本许可冲突之处，以本许可为准。
#         详细的授权流程，请联系 public@ricequant.com 获取。

import os
from copy import deepcopy
from importlib import import_module
from typing import Mapping

import six


def deep_update(from_dict, to_dict):
    for (key, value) in from_dict.items():
        if (key in to_dict.keys() and isinstance(to_dict[key], Mapping) and isinstance(value, Mapping)):
            deep_update(value, to_dict[key])
        else:
            to_dict[key] = value


def import_tests(dir, module):
    strategies = {}

    for file in os.listdir(dir):
        file_path = os.path.join(dir, file)
        if os.path.isdir(file_path):
            if not file.startswith("__"):
                strategies.update(import_tests(file_path, ".".join((module, file))))
        else:
            if file.endswith(".py") and file.startswith("test"):
                m = import_module(".".join((module, file[:-3])), "tests.api_tests")
                default_config = m.__dict__.pop("__config__", {})
                for obj in six.itervalues(m.__dict__):
                    if hasattr(obj, "__call__") and getattr(obj, "__name__", "").startswith("test"):
                        strategy_locals = obj()
                        custom_config = strategy_locals.pop("__config__", {})
                        config = deepcopy(default_config)
                        if custom_config:
                            deep_update(custom_config, config)
                        strategy_locals["config"] = config
                        strategy_locals["name"] = obj.__name__
                        strategies[".".join((module, file[:-3], obj.__name__))] = strategy_locals
    return strategies


strategies = list(six.itervalues(import_tests(os.path.dirname(__file__), "")))



# -*- coding: utf-8 -*-
# 版权所有 2019 深圳米筐科技有限公司（下称“米筐科技”）
#
# 除非遵守当前许可，否则不得使用本软件。
#
#     * 非商业用途（非商业用途指个人出于非商业目的使用本软件，或者高校、研究所等非营利机构出于教育、科研等目的使用本软件）：
#         遵守 Apache License 2.0（下称“Apache 2.0 许可”），您可以在以下位置获得 Apache 2.0 许可的副本：http://www.apache.org/licenses/LICENSE-2.0。
#         除非法律有要求或以书面形式达成协议，否则本软件分发时需保持当前许可“原样”不变，且不得附加任何条件。
#
#     * 商业用途（商业用途指个人出于任何商业目的使用本软件，或者法人或其他组织出于任何目的使用本软件）：
#         未经米筐科技授权，任何个人不得出于任何商业目的使用本软件（包括但不限于向第三方提供、销售、出租、出借、转让本软件、本软件的衍生产品、引用或借鉴了本软件功能或源代码的产品或服务），任何法人或其他组织不得出于任何目的使用本软件，否则米筐科技有权追究相应的知识产权侵权责任。
#         在此前提下，对本软件的使用同样需要遵守 Apache 2.0 许可，Apache 2.0 许可与本许可冲突之处，以本许可为准。
#         详细的授权流程，请联系 public@ricequant.com 获取。

import os
import datetime
import pickle
import tempfile

from rqalpha.utils.testing import DataProxyFixture, RQAlphaTestCase
from rqalpha.data.bundle import AutomaticUpdateBundle
from rqalpha.data.trading_dates_mixin import TradingDatesMixin


class AutomaticUpdateBundleTestCase(DataProxyFixture, RQAlphaTestCase):
    def __init__(self, *args, **kwargs):
        super(AutomaticUpdateBundleTestCase, self).__init__(*args, **kwargs)
        self._path = tempfile.TemporaryDirectory().name

    def init_fixture(self):
        super(AutomaticUpdateBundleTestCase, self).init_fixture()
        self._auto_update_bundle_module = AutomaticUpdateBundle(
            path=self._path,
            filename="open_auction_volume.h5",
            api=self._mock_get_open_auction_info,
            fields=['volume'],
            end_date=datetime.date(2024, 2, 28),
        )
        trading_dates_mixin = TradingDatesMixin(self.base_data_source.get_trading_calendars())
        self.base_data_source.batch_get_trading_date = trading_dates_mixin.batch_get_trading_date
        self.base_data_source.get_next_trading_date = trading_dates_mixin.get_next_trading_date

    def _mock_get_open_auction_info(self, order_book_id, *args, **kwargs):
        df = pickle.loads(open(
            os.path.join(os.path.dirname(__file__), "mock_data/mock_open_auction_info.pkl"), "rb"
        ).read())
        df = df.loc[order_book_id].reset_index()
        df['order_book_id'] = order_book_id
        df = df.set_index(["order_book_id", "datetime"])
        return df
    
    def _mock_get_open_auction_volume(self, instrument, dt):
        # type: (Instrument, datetime.date) -> float
        data = self._auto_update_bundle_module.get_data(instrument, dt)
        if data is None:
            volume = 0
        else:
            volume = 0 if len(data) == 0 else data['volume']
        return volume

    def test_auto_update_bundle(self):
        s_volume = self._mock_get_open_auction_volume(self.env.get_instrument("000001.XSHE"), datetime.date(2023, 12, 28))
        f_volume = self._mock_get_open_auction_volume(self.env.get_instrument("A2401"), datetime.date(2023, 12, 28))
        assert os.path.exists(os.path.join(self._path, "open_auction_volume.h5")) == True

        pickle_data = pickle.loads(open(
            os.path.join(os.path.dirname(__file__), "mock_data/mock_open_auction_info.pkl"), "rb"
        ).read())

        s_df = pickle_data.loc["000001.XSHE"]
        assert s_volume == s_df[s_df.index.date == datetime.date(2023, 12, 28)].volume[0]
        f_df = pickle_data.loc['A2401']
        # 期货由于有夜盘，open_auction_info 的时间为前一个交易日的晚上，即从 pickle 文件中获取时，date 应该提前一个交易日
        assert f_volume == f_df[f_df.index.date == datetime.date(2023, 12, 27)].volume[0]
        



# -*- coding: utf-8 -*-
# 版权所有 2019 深圳米筐科技有限公司（下称“米筐科技”）
#
# 除非遵守当前许可，否则不得使用本软件。
#
#     * 非商业用途（非商业用途指个人出于非商业目的使用本软件，或者高校、研究所等非营利机构出于教育、科研等目的使用本软件）：
#         遵守 Apache License 2.0（下称“Apache 2.0 许可”），
#         您可以在以下位置获得 Apache 2.0 许可的副本：http://www.apache.org/licenses/LICENSE-2.0。
#         除非法律有要求或以书面形式达成协议，否则本软件分发时需保持当前许可“原样”不变，且不得附加任何条件。
#
#     * 商业用途（商业用途指个人出于任何商业目的使用本软件，或者法人或其他组织出于任何目的使用本软件）：
#         未经米筐科技授权，任何个人不得出于任何商业目的使用本软件（包括但不限于向第三方提供、销售、出租、出借、转让本软件、
#         本软件的衍生产品、引用或借鉴了本软件功能或源代码的产品或服务），任何法人或其他组织不得出于任何目的使用本软件，
#         否则米筐科技有权追究相应的知识产权侵权责任。
#         在此前提下，对本软件的使用同样需要遵守 Apache 2.0 许可，Apache 2.0 许可与本许可冲突之处，以本许可为准。
#         详细的授权流程，请联系 public@ricequant.com 获取。

from typing import Any

from rqalpha.utils import INST_TYPE_IN_STOCK_ACCOUNT
from rqalpha.utils.logger import user_system_log
from rqalpha.interface import AbstractMod
from rqalpha.const import DEFAULT_ACCOUNT_TYPE, INSTRUMENT_TYPE
from rqalpha.environment import Environment

from .position_model import StockPosition
from .position_validator import PositionValidator
from .component_validator import MarginComponentValidator
from .validator import MarginInstrumentValidator


class AccountMod(AbstractMod):

    def start_up(self, env, mod_config):
        # type: (Environment, Any) -> None

        StockPosition.dividend_reinvestment = mod_config.dividend_reinvestment
        StockPosition.cash_return_by_stock_delisted = mod_config.cash_return_by_stock_delisted
        StockPosition.t_plus_enabled = mod_config.stock_t1

        pos_validator = PositionValidator()
        if mod_config.validate_future_position:
            env.add_frontend_validator(pos_validator, INSTRUMENT_TYPE.FUTURE)
        if mod_config.validate_stock_position:
            for ins_type in INST_TYPE_IN_STOCK_ACCOUNT:
                env.add_frontend_validator(pos_validator, ins_type)

        if not isinstance(mod_config.financing_rate, (int, float)):
            raise ValueError("sys_accounts financing_rate must number")
        elif mod_config.financing_rate < 0:
            raise ValueError("sys_accounts financing_rate must >= 0")

        futures_settlement_price_types = ["settlement", "close"]
        if mod_config.futures_settlement_price_type not in futures_settlement_price_types:
            raise ValueError("sys_accounts futures_settlement_price_type in {}".format(futures_settlement_price_types))

        if DEFAULT_ACCOUNT_TYPE.FUTURE in env.config.base.accounts:
            # 注入期货API
            # noinspection PyUnresolvedReferences
            from .api import api_future
        if DEFAULT_ACCOUNT_TYPE.STOCK in env.config.base.accounts:
            # 注入股票API
            # noinspection PyUnresolvedReferences
            from .api import api_stock

            # 启动融资股票池限制
            if mod_config.financing_stocks_restriction_enabled:
                try:
                    import rqdatac
                    if rqdatac.initialized():
                        com_validator = MarginComponentValidator(margin_type="all")
                        ins_validator = MarginInstrumentValidator()
                        # 融资股票池验证
                        env.add_frontend_validator(com_validator, INSTRUMENT_TYPE.CS)
                        env.add_frontend_validator(com_validator, INSTRUMENT_TYPE.ETF)
                        # 下单标的验证
                        env.add_frontend_validator(ins_validator, INSTRUMENT_TYPE.LOF)
                        env.add_frontend_validator(ins_validator, INSTRUMENT_TYPE.CONVERTIBLE)
                        env.add_frontend_validator(ins_validator, INSTRUMENT_TYPE.INDX)
                        env.add_frontend_validator(ins_validator, INSTRUMENT_TYPE.PUBLIC_FUND)
                        env.add_frontend_validator(ins_validator, INSTRUMENT_TYPE.REITs)
                    else:
                        user_system_log.warn("rqdatac not init, not support financing stocks restriction.")
                except ImportError:
                    user_system_log.warn("rqdatac not install, not support financing stocks restriction.")

    def tear_down(self, code, exception=None):
        pass



import os
from typing import Generator, Any, Optional, Callable, Tuple
import queue
import multiprocessing

import click

from concurrent.futures import _base
# noinspection PyUnresolvedReferences
from concurrent.futures.process import ProcessPoolExecutor, _ExceptionWithTraceback, _ResultItem


def _process_worker(call_queue, result_queue, progress_queue, initializer, initargs):
    if initializer is not None:
        try:
            initializer(*initargs)
        except BaseException:
            _base.LOGGER.critical('Exception in initializer:', exc_info=True)
            return
    while True:
        call_item = call_queue.get(block=True)
        if call_item is None:
            result_queue.put(os.getpid())
            return
        try:
            r = call_item.fn(*call_item.args, **call_item.kwargs)
            if isinstance(call_item.fn, ProgressedTask):
                for step in r:
                    progress_queue.put(step)
                r = None
            else:
                progress_queue.put(1)
        except BaseException as e:
            exc = _ExceptionWithTraceback(e, e.__traceback__)
            result_queue.put(_ResultItem(call_item.work_id, exception=exc))
        else:
            result_queue.put(_ResultItem(call_item.work_id, result=r))


class ProgressedProcessPoolExecutor(ProcessPoolExecutor):
    def __init__(self, max_workers=None, initializer=None, initargs=()):
        # type: (Optional[int], Optional[Callable], Optional[Tuple]) -> None
        super(ProgressedProcessPoolExecutor, self).__init__(max_workers)
        self._initializer = initializer
        self._initargs = initargs
        self._progress_queue = multiprocessing.Queue()
        self._futures = []
        self._total_steps = 0

    def _adjust_process_count(self):
        # noinspection PyUnresolvedReferences
        for _ in range(len(self._processes), self._max_workers):
            self._spawn_process()

    def _spawn_process(self):
        p = multiprocessing.Process(
            target=_process_worker,
            args=(self._call_queue, self._result_queue, self._progress_queue, self._initializer, self._initargs)
        )
        p.start()
        # noinspection PyUnresolvedReferences
        self._processes[p.pid] = p

    def submit(self, fn, *args, **kwargs):
        if isinstance(fn, ProgressedTask):
            # noinspection PyUnresolvedReferences
            self._total_steps += fn.total_steps
        else:
            self._total_steps += 1
        f = super(ProgressedProcessPoolExecutor, self).submit(fn, *args, **kwargs)
        self._futures.append(f)
        return f

    def shutdown(self, wait=True):
        if not wait:
            return super(ProgressedProcessPoolExecutor, self).shutdown(wait)
        progress_bar = click.progressbar(length=self._total_steps, show_eta=False)
        finish = False
        while True:
            for fut in self._futures:
                if fut.running():
                    break
            else:
                finish = True
            while True:
                try:
                    step = self._progress_queue.get(timeout=0.5)
                except queue.Empty:
                    break
                else:
                    progress_bar.update(step)

            if finish:
                break

        progress_bar.render_finish()
        super(ProgressedProcessPoolExecutor, self).shutdown(True)
        for fut in self._futures:
            if fut.exception():
                raise fut.exception()


class ProgressedTask:
    @property
    def total_steps(self):
        # type: () -> int
        raise NotImplementedError

    def __call__(self, *args, **kwargs):
        # type: (*Any, **Any) -> Generator
        raise NotImplementedError



# -*- coding: utf-8 -*-
# 版权所有 2019 深圳米筐科技有限公司（下称"米筐科技"）
#
# 除非遵守当前许可，否则不得使用本软件。
#
#     * 非商业用途（非商业用途指个人出于非商业目的使用本软件，或者高校、研究所等非营利机构出于教育、科研等目的使用本软件）：
#         遵守 Apache License 2.0（下称"Apache 2.0 许可"），
#         您可以在以下位置获得 Apache 2.0 许可的副本：http://www.apache.org/licenses/LICENSE-2.0。
#         除非法律有要求或以书面形式达成协议，否则本软件分发时需保持当前许可"原样"不变，且不得附加任何条件。
#
#     * 商业用途（商业用途指个人出于任何商业目的使用本软件，或者法人或其他组织出于任何目的使用本软件）：
#         未经米筐科技授权，任何个人不得出于任何商业目的使用本软件（包括但不限于向第三方提供、销售、出租、出借、转让本软件、
#         本软件的衍生产品、引用或借鉴了本软件功能或源代码的产品或服务），任何法人或其他组织不得出于任何目的使用本软件，
#         否则米筐科技有权追究相应的知识产权侵权责任。
#         在此前提下，对本软件的使用同样需要遵守 Apache 2.0 许可，Apache 2.0 许可与本许可冲突之处，以本许可为准。
#         详细的授权流程，请联系 public@ricequant.com 获取。

from numpy.testing import assert_equal
from rqalpha.apis import *

__config__ = {
    "base": {
        "start_date": "2016-03-07",
        "end_date": "2016-03-10",
        "frequency": "1d",
        "accounts": {
            "stock": 1000000,
            "future": 1000000
        }
    },
    "extra": {
        "log_level": "error",
    },
    "mod": {
        "sys_progress": {
            "enabled": True,
            "show": True,
        },
    },
}


def test_stock_position_queue_open_close():
    """
    测试股票开仓和平仓时的 position_queue
    """
    def init(context):
        context.s = "000001.XSHE"
        context.counter = 0

    def handle_bar(context, bar_dict):
        context.counter += 1

        if context.counter == 1:
            # 第一天买入1000股
            order_shares(context.s, 1000)
            position = get_position(context.s)
            assert_equal(position.quantity, 1000)

            # 检查 position_queue
            queue = position.position_queue
            assert_equal(len(queue), 1)
            assert_equal(queue[0][0], context.now.date())
            assert_equal(queue[0][1], 1000)

        elif context.counter == 2:
            # 第二天再买入500股
            order_shares(context.s, 500)
            position = get_position(context.s)
            assert_equal(position.quantity, 1500)

            # 检查 position_queue
            queue = position.position_queue
            assert_equal(len(queue), 2)
            assert_equal(queue[0][0], date(2016, 3, 7))
            assert_equal(queue[0][1], 1000)
            assert_equal(queue[1][0], context.now.date())
            assert_equal(queue[1][1], 500)

        elif context.counter == 3:
            # 第三天卖出800股，应该先卖出第一天买入的部分
            order_shares(context.s, -800)
            position = get_position(context.s)
            assert_equal(position.quantity, 700)

            # 检查 position_queue
            queue = position.position_queue
            assert_equal(len(queue), 2)
            assert_equal(queue[0][0], date(2016, 3, 7))
            assert_equal(queue[0][1], 200)  # 1000 - 800 = 200
            assert_equal(queue[1][0], date(2016, 3, 8))
            assert_equal(queue[1][1], 500)

        elif context.counter == 4:
            # 第四天全部卖出
            order_shares(context.s, -700)
            position = get_position(context.s)
            assert_equal(position.quantity, 0)

            # 检查 position_queue
            queue = position.position_queue
            assert_equal(len(queue), 0)  # 应该为空

    return locals()


def test_stock_position_queue_short_selling():
    """
    测试股票卖空时的 position_queue
    注意：此测试需要在支持卖空的环境中运行，否则会跳过测试
    """
    __config__ = {
        "base": {
            "start_date": "2016-03-07",
            "end_date": "2016-03-10",
            "accounts": {
                "stock": 1000000
            }
        },
        "mod": {
            "sys_accounts": {
                "stock_t0": True,  # 启用T+0交易以便测试卖空
                "future_forced_liquidation": True,
                "validate_stock_position": False  # 禁用股票仓位检查以启用卖空
            }
        }
    }

    def init(context):
        context.s = "000001.XSHE"
        context.counter = 0

    def handle_bar(context, bar_dict):
        context.counter += 1

        if context.counter == 1:
            # 尝试卖空1000股
            order_shares(context.s, -1000)
            position = get_position(context.s)

            # 检查 position_queue
            queue = position.position_queue
            assert_equal(len(queue), 1)
            assert_equal(queue[0][0], context.now.date())
            assert_equal(queue[0][1], -1000)

        elif context.counter == 2:
            # 第二天买入500股平仓一部分
            order_shares(context.s, 500)
            position = get_position(context.s)

            # 检查 position_queue
            queue = position.position_queue
            assert_equal(len(queue), 1)
            assert_equal(queue[0][0], date(2016, 3, 7))
            assert_equal(queue[0][1], -500)

        elif context.counter == 3:
            # 第三天买入足够股数平仓
            current_position = get_position(context.s)
            if current_position.quantity < 0:
                order_shares(context.s, -current_position.quantity + 500)
                position = get_position(context.s)

                # 检查 position_queue
                queue = position.position_queue
                assert_equal(len(queue), 1)
                assert_equal(queue[0][0], context.now.date())
                assert_equal(queue[0][1], 500)  # 应该变成+500

    return locals()


def test_stock_position_queue_split():
    """
    测试股票拆分时的 position_queue
    """
    __config__ = {
        "base": {
            "start_date": "2024-06-17",
            "end_date": "2024-06-19"
        },
        "mod": {
            "sys_accounts": {
                "validate_stock_position": False  # 禁用股票仓位检查以启用卖空
            }
        }
    }

    def init(context):
        context.s = "688032.XSHG"
        context.counter = 0

    def handle_bar(context, bar_dict):
        context.counter += 1

        if context.counter == 1:
            # 第一天买入1000股
            order_shares(context.s, -303)
            position = get_position(context.s)
            assert_equal(position.quantity, -303)

            # 检查 position_queue
            queue = position.position_queue
            assert_equal(len(queue), 1)
            assert_equal(queue[0][0], context.now.date())
            assert_equal(queue[0][1], -303)

        elif context.counter == 2:
            order_shares(context.s, -415)
        elif context.counter == 3:
            position = get_position(context.s)
            assert_equal(position.quantity, -1070)

            # 检查 position_queue
            queue = position.position_queue
            assert_equal(len(queue), 2)
            assert_equal(queue[0][0], date(2024, 6, 17))
            assert_equal(queue[0][1] + queue[1][1], -1070)

    return locals()


def test_stock_position_queue_delist():
    """
    测试股票退市时的 position_queue
    """
    __config__ = {
        "base": {
            "start_date": "2018-12-25",
            "end_date": "2019-01-05"
        }
    }

    def init(context):
        context.s = "000979.XSHE"  # 这只股票在2018-12-28退市
        context.counter = 0

    def handle_bar(context, bar_dict):
        context.counter += 1

        if context.counter == 1:
            # 买入股票
            order_shares(context.s, 1000)
            position = get_position(context.s)
            assert_equal(position.quantity, 1000)

            # 检查 position_queue
            queue = position.position_queue
            assert_equal(len(queue), 1)
            assert_equal(queue[0][0], context.now.date())
            assert_equal(queue[0][1], 1000)

        elif context.counter > 3:  # 退市后
            # 退市后应该没有持仓了
            position = get_position(context.s)
            assert_equal(position.quantity, 0)

            # 检查 position_queue
            queue = position.position_queue
            assert_equal(len(queue), 0)  # 应该为空

    return locals()


def test_future_position_queue_close_today():
    """
    测试期货平今仓时的 position_queue
    """
    __config__ = {
        "base": {
            "start_date": "2016-03-07",
            "end_date": "2016-03-08",
            "accounts": {
                "future": 1000000
            }
        }
    }

    def init(context):
        context.f = "IF1603"  # 选择一个期货合约
        context.initial_position = 0
        context.initial_queue_length = 0
        subscribe(context.f)

    def handle_bar(context, bar_dict):
        # 检查初始持仓
        initial_position = get_position(context.f, POSITION_DIRECTION.LONG)
        context.initial_position = initial_position.quantity
        context.initial_queue_length = len(initial_position.position_queue)

        # 买入开仓2手
        buy_open(context.f, 2)
        position = get_position(context.f, POSITION_DIRECTION.LONG)
        expected_quantity = context.initial_position + 2
        assert_equal(position.quantity, expected_quantity)

        # 检查 position_queue 总数量
        queue = position.position_queue
        total_quantity_in_queue = sum(q[1] for q in queue)
        assert_equal(total_quantity_in_queue, expected_quantity)

        # 检查今日持仓
        today_position = 0
        for q in queue:
            if q[0] == context.now.date():
                today_position += q[1]
        assert today_position > 0, "今日应该有持仓"

        # 平今仓1手
        sell_close(context.f, 1, close_today=True)
        position = get_position(context.f, POSITION_DIRECTION.LONG)
        expected_quantity_after_close = expected_quantity - 1
        assert_equal(position.quantity, expected_quantity_after_close)

        # 检查 position_queue 总数量
        queue = position.position_queue
        total_quantity_in_queue = sum(q[1] for q in queue)
        assert_equal(total_quantity_in_queue, expected_quantity_after_close)

        # 检查今日持仓是否减少
        today_position_after_close = 0
        for q in queue:
            if q[0] == context.now.date():
                today_position_after_close += q[1]

        # 如果今日有持仓，应该比之前少1手
        if today_position_after_close > 0:
            assert_equal(today_position - today_position_after_close, 1)

    return locals()


def test_comprehensive_position_queue():
    """
    综合测试 position_queue 的各种情况
    """
    __config__ = {
        "base": {
            "start_date": "2016-03-07",
            "end_date": "2016-03-11",
            "accounts": {
                "stock": 1000000,
                "future": 1000000
            }
        }
    }

    def init(context):
        context.stock = "000001.XSHE"
        context.future = "IF1603"
        context.counter = 0
        subscribe(context.future)

    def handle_bar(context, bar_dict):
        context.counter += 1

        if context.counter == 1:
            # 第一天：股票买入1000股，期货买入开仓2手
            order_shares(context.stock, 1000)
            buy_open(context.future, 2)

            # 检查股票 position_queue
            stock_position = get_position(context.stock)
            stock_queue = stock_position.position_queue
            assert_equal(len(stock_queue), 1)
            assert_equal(stock_queue[0][0], context.now.date())
            assert_equal(stock_queue[0][1], 1000)

            # 检查期货 position_queue
            future_position = get_position(context.future, POSITION_DIRECTION.LONG)
            future_queue = future_position.position_queue
            assert_equal(len(future_queue), 1)
            assert_equal(future_queue[0][0], context.now.date())
            assert_equal(future_queue[0][1], 2)

        elif context.counter == 2:
            # 第二天：股票卖出500股，期货平今仓1手
            order_shares(context.stock, -500)
            sell_close(context.future, 1, close_today=False)  # 不强制平今

            # 检查股票 position_queue
            stock_position = get_position(context.stock)
            stock_queue = stock_position.position_queue
            assert_equal(len(stock_queue), 1)
            assert_equal(stock_queue[0][0], date(2016, 3, 7))
            assert_equal(stock_queue[0][1], 500)  # 1000 - 500 = 500

            # 检查期货 position_queue
            future_position = get_position(context.future, POSITION_DIRECTION.LONG)
            future_queue = future_position.position_queue
            assert_equal(len(future_queue), 1)
            assert_equal(future_queue[0][0], date(2016, 3, 7))
            assert_equal(future_queue[0][1], 1)  # 2 - 1 = 1

        elif context.counter == 3:
            # 第三天：股票买入200股，期货卖出开仓1手
            order_shares(context.stock, 200)
            sell_open(context.future, 1)

            # 检查股票 position_queue
            stock_position = get_position(context.stock)
            stock_queue = stock_position.position_queue
            assert_equal(len(stock_queue), 2)
            assert_equal(stock_queue[0][0], date(2016, 3, 7))
            assert_equal(stock_queue[0][1], 500)
            assert_equal(stock_queue[1][0], context.now.date())
            assert_equal(stock_queue[1][1], 200)

            # 检查期货多头 position_queue
            future_long_position = get_position(context.future, POSITION_DIRECTION.LONG)
            future_long_queue = future_long_position.position_queue
            assert_equal(len(future_long_queue), 1)
            assert_equal(future_long_queue[0][0], date(2016, 3, 7))
            assert_equal(future_long_queue[0][1], 1)

            # 检查期货空头 position_queue
            future_short_position = get_position(context.future, POSITION_DIRECTION.SHORT)
            future_short_queue = future_short_position.position_queue
            assert_equal(len(future_short_queue), 1)
            assert_equal(future_short_queue[0][0], context.now.date())
            assert_equal(future_short_queue[0][1], 1)

        elif context.counter == 4:
            # 第四天：股票卖出全部，期货平掉所有仓位
            order_shares(context.stock, -700)
            sell_close(context.future, 1)  # 平掉多头
            buy_close(context.future, 1)   # 平掉空头

            # 检查股票 position_queue
            stock_position = get_position(context.stock)
            stock_queue = stock_position.position_queue
            assert_equal(len(stock_queue), 0)  # 应该为空

            # 检查期货多头 position_queue
            future_long_position = get_position(context.future, POSITION_DIRECTION.LONG)
            future_long_queue = future_long_position.position_queue
            assert_equal(len(future_long_queue), 0)  # 应该为空

            # 检查期货空头 position_queue
            future_short_position = get_position(context.future, POSITION_DIRECTION.SHORT)
            future_short_queue = future_short_position.position_queue
            assert_equal(len(future_short_queue), 0)  # 应该为空

    return locals()



# -*- coding: utf-8 -*-
# 版权所有 2019 深圳米筐科技有限公司（下称“米筐科技”）
#
# 除非遵守当前许可，否则不得使用本软件。
#
#     * 非商业用途（非商业用途指个人出于非商业目的使用本软件，或者高校、研究所等非营利机构出于教育、科研等目的使用本软件）：
#         遵守 Apache License 2.0（下称“Apache 2.0 许可”），您可以在以下位置获得 Apache 2.0 许可的副本：http://www.apache.org/licenses/LICENSE-2.0。
#         除非法律有要求或以书面形式达成协议，否则本软件分发时需保持当前许可“原样”不变，且不得附加任何条件。
#
#     * 商业用途（商业用途指个人出于任何商业目的使用本软件，或者法人或其他组织出于任何目的使用本软件）：
#         未经米筐科技授权，任何个人不得出于任何商业目的使用本软件（包括但不限于向第三方提供、销售、出租、出借、转让本软件、本软件的衍生产品、引用或借鉴了本软件功能或源代码的产品或服务），任何法人或其他组织不得出于任何目的使用本软件，否则米筐科技有权追究相应的知识产权侵权责任。
#         在此前提下，对本软件的使用同样需要遵守 Apache 2.0 许可，Apache 2.0 许可与本许可冲突之处，以本许可为准。
#         详细的授权流程，请联系 public@ricequant.com 获取。
from typing import Optional

from rqalpha.interface import AbstractFrontendValidator
from rqalpha.const import ORDER_TYPE, POSITION_EFFECT
from rqalpha.model.order import Order
from rqalpha.portfolio.account import Account
from rqalpha.utils.logger import user_system_log

from rqalpha.utils.i18n import gettext as _


class PriceValidator(AbstractFrontendValidator):
    def __init__(self, env):
        self._env = env
    
    def validate_submission(self, order: Order, account: Optional[Account] = None) -> Optional[str]:
        if (order.type != ORDER_TYPE.LIMIT) or (order.position_effect == POSITION_EFFECT.EXERCISE):
            return None
        
        # FIXME: it may be better to round price in data source
        limit_up = round(self._env.price_board.get_limit_up(order.order_book_id), 4)
        if order.price > limit_up:
            reason = _(
                "Order Creation Failed: limit order price {limit_price} is higher "
                "than limit up {limit_up}, order_book_id={order_book_id}"
            ).format(
                order_book_id=order.order_book_id,
                limit_price=order.price,
                limit_up=limit_up
            )
            return reason

        limit_down = round(self._env.price_board.get_limit_down(order.order_book_id), 4)
        if order.price < limit_down:
            reason = _(
                "Order Creation Failed: limit order price {limit_price} is lower "
                "than limit down {limit_down}, order_book_id={order_book_id}"
            ).format(
                order_book_id=order.order_book_id,
                limit_price=order.price,
                limit_down=limit_down
            )
            return reason
        return None

    def validate_cancellation(self, order: Order, account: Optional[Account] = None) -> Optional[str]:
        return None